import numpy as np
from karabo.bound import KARABO_CLASSINFO, VECTOR_STRING_ELEMENT
from karabo.common.states import State

from . import utils
from ._version import version as deviceVersion
from .base_correction import BaseCorrection, add_correction_step_schema
from .calcat_utils import DsscCalcatFriend, DsscConstants
from .dssc_gpu import CorrectionFlags, DsscGpuRunner


@KARABO_CLASSINFO("DsscCorrection", deviceVersion)
class DsscCorrection(BaseCorrection):
    # subclass *must* set these attributes
    _correction_flag_class = CorrectionFlags
    _correction_field_names = (("offset", CorrectionFlags.OFFSET),)
    _kernel_runner_class = DsscGpuRunner
    _calcat_friend_class = DsscCalcatFriend
    _constant_enum_class = DsscConstants
    _managed_keys = BaseCorrection._managed_keys.copy()

    @staticmethod
    def expectedParameters(expected):
        super(DsscCorrection, DsscCorrection).expectedParameters(expected)
        expected.setDefaultValue("dataFormat.memoryCells", 400)
        expected.setDefaultValue("preview.selectionMode", "pulse")
        DsscCalcatFriend.add_schema(expected, DsscCorrection._managed_keys)
        add_correction_step_schema(
            expected,
            DsscCorrection._managed_keys,
            DsscCorrection._correction_field_names,
        )
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("managedKeys")
            .assignmentOptional()
            .defaultValue(list(DsscCorrection._managed_keys))
            .commit()
        )

    @property
    def input_data_shape(self):
        return (
            self.get("dataFormat.memoryCells"),
            1,
            self.get("dataFormat.pixelsY"),
            self.get("dataFormat.pixelsX"),
        )

    def __init__(self, config):
        super().__init__(config)
        self.updateState(State.ON)

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        do_generate_preview,
    ):
        pulse_table = np.squeeze(data_hash.get("image.pulseId"))
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            self.kernel_runner.load_data(image_data)
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.load_cell_table(cell_table)
        self.kernel_runner.correct(self._correction_flag_enabled)
        self.kernel_runner.reshape(
            output_order=self._schema_cache["dataFormat.outputAxisOrder"],
            out=buffer_array,
        )
        if do_generate_preview:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ) = utils.pick_frame_index(
                self._schema_cache["preview.selectionMode"],
                self._schema_cache["preview.index"],
                cell_table,
                pulse_table,
                warn_func=self.log_status_warn,
            )
            preview_raw, preview_corrected = self.kernel_runner.compute_previews(
                preview_slice_index,
            )

        data_hash.set("image.data", buffer_handle)
        data_hash.set("image.cellId", cell_table[:, np.newaxis])
        data_hash.set("image.pulseId", pulse_table[:, np.newaxis])
        data_hash.set("calngShmemPaths", ["image.data"])
        self._write_output(data_hash, metadata)
        if do_generate_preview:
            self._write_combiner_previews(
                (
                    ("preview.outputRaw", preview_raw),
                    ("preview.outputCorrected", preview_corrected),
                ),
                train_id,
                source,
            )

    def _load_constant_to_runner(self, constant, constant_data):
        assert constant is DsscConstants.Offset
        self.kernel_runner.load_offset_map(constant_data)
        if not self.get("corrections.offset.available"):
            self.set("corrections.offset.available", True)

        self._update_correction_flags()
        self.log_status_info(f"Done loading {constant.name} to GPU")
