import pathlib
import threading
from timeit import default_timer

import dateutil.parser
import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    INPUT_CHANNEL,
    INT32_ELEMENT,
    INT64_ELEMENT,
    KARABO_CLASSINFO,
    NDARRAY_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    VECTOR_STRING_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    MetricPrefix,
    PythonDevice,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES
from karabo.common.states import State
from karabo import version as karaboVersion
from pkg_resources import parse_version

from . import scenes, shmem_utils, utils
from ._version import version as deviceVersion

PROCESSING_STATE_TIMEOUT = 10

preview_schema = Schema()
(
    NODE_ELEMENT(preview_schema).key("image").commit(),
    NDARRAY_ELEMENT(preview_schema).key("image.data").dtype("FLOAT").commit(),
    UINT64_ELEMENT(preview_schema)
    .key("image.trainId")
    .displayedName("Train ID")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
)

# TODO: trim output schema / adapt to specific detectors
# currently: based on snapshot of actual output reusing AGIPD hash
output_schema = Schema()
(
    NODE_ELEMENT(output_schema).key("image").commit(),
    STRING_ELEMENT(output_schema)
    .key("image.data")
    .assignmentOptional()
    .defaultValue("")
    .commit(),
    NDARRAY_ELEMENT(output_schema).key("image.length").dtype("UINT32").commit(),
    NDARRAY_ELEMENT(output_schema).key("image.cellId").dtype("UINT16").commit(),
    NDARRAY_ELEMENT(output_schema).key("image.pulseId").dtype("UINT64").commit(),
    NDARRAY_ELEMENT(output_schema).key("image.status").commit(),
    NDARRAY_ELEMENT(output_schema).key("image.trainId").dtype("UINT64").commit(),
    VECTOR_STRING_ELEMENT(output_schema)
    .key("calngShmemPaths")
    .assignmentOptional()
    .defaultValue(["image.data"])
    .commit(),
    NODE_ELEMENT(output_schema).key("metadata").commit(),
    STRING_ELEMENT(output_schema)
    .key("metadata.source")
    .assignmentOptional()
    .defaultValue("")
    .commit(),
    NODE_ELEMENT(output_schema).key("metadata.timestamp").commit(),
    INT32_ELEMENT(output_schema)
    .key("metadata.timestamp.tid")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    NODE_ELEMENT(output_schema).key("header").commit(),
    INT32_ELEMENT(output_schema)
    .key("header.minorTrainFormatVersion")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    INT32_ELEMENT(output_schema)
    .key("header.majorTrainFormatVersion")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    INT32_ELEMENT(output_schema)
    .key("header.trainId")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    INT64_ELEMENT(output_schema)
    .key("header.linkId")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    INT64_ELEMENT(output_schema)
    .key("header.dataId")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    INT64_ELEMENT(output_schema)
    .key("header.pulseCount")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    NDARRAY_ELEMENT(output_schema).key("header.reserved").commit(),
    NDARRAY_ELEMENT(output_schema).key("header.magicNumberBegin").commit(),
    NODE_ELEMENT(output_schema).key("detector").commit(),
    INT32_ELEMENT(output_schema)
    .key("detector.trainId")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    NDARRAY_ELEMENT(output_schema).key("detector.data").commit(),
    NODE_ELEMENT(output_schema).key("trailer").commit(),
    NDARRAY_ELEMENT(output_schema).key("trailer.checksum").commit(),
    NDARRAY_ELEMENT(output_schema).key("trailer.magicNumberEnd").commit(),
    INT32_ELEMENT(output_schema)
    .key("trailer.status")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
    INT32_ELEMENT(output_schema)
    .key("trailer.trainId")
    .assignmentOptional()
    .defaultValue(0)
    .commit(),
)


@KARABO_CLASSINFO("BaseCorrection", deviceVersion)
class BaseCorrection(PythonDevice):
    _correction_flag_class = None  # subclass must set (ex.: dssc_gpu.CorrectionFlags)
    _kernel_runner_class = None  # subclass must set (ex.: dssc_gpu.DsscGpuRunner)
    _kernel_runner_init_args = {}  # optional extra args for runner
    _managed_keys = {
        "outputShmemBufferSize",
        "dataFormat.outputAxisOrder",
        "dataFormat.outputImageDtype",
        "dataFormat.overrideInputAxisOrder",
        "frameFilter",
        "preview.enable",
        "preview.index",
        "preview.selectionMode",
        "preview.trainIdModulo",
        "loadMostRecentConstants",
    }  # subclass can extend this, /must/ put it in schema as managedKeys
    _schema_cache_fields = {
        "doAnything",
        "constantParameters.memoryCells",
        "dataFormat.filteredFrames",
        "dataFormat.memoryCells",
        "dataFormat.pixelsX",
        "dataFormat.pixelsY",
        "dataFormat.outputAxisOrder",
        "dataFormat.overrideInputAxisOrder",
        "preview.enable",
        "preview.index",
        "preview.selectionMode",
        "preview.trainIdModulo",
        "processingStateTimeout",
        "state",
    }  # subclass should be aware of cache, but does not need to extend

    def _load_constant_to_runner(constant_name, constant_data):
        """Subclass must define how to process constants into correction maps and store
        into appropriate buffers in (GPU or main) memory."""
        raise NotImplementedError()

    @property
    def input_data_shape(self):
        """Subclass must define expected input data shape in terms of dataFormat.{
        memoryCells,pixelsX,pixelsY} and any other axes."""
        raise NotImplementedError()

    @property
    def output_data_shape(self):
        """Shape of corrected image data sent on dataOutput. Depends on data format
        parameters pixels x / y, and number of cells (optionally after frame filter)."""
        axis_lengths = {
            "x": self._schema_cache["dataFormat.pixelsX"],
            "y": self._schema_cache["dataFormat.pixelsY"],
            "c": self._schema_cache["dataFormat.filteredFrames"],
        }
        return tuple(
            axis_lengths[axis]
            for axis in self._schema_cache["dataFormat.outputAxisOrder"]
        )

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        do_generate_preview,
    ):
        """Subclass must define data processing (presumably using the kernel runner).
        Will be called by input_handler, which will take care of some common checks "
        "and extracting the parameters given to process_data."""
        raise NotImplementedError()

    @staticmethod
    def expectedParameters(expected):
        (
            INPUT_CHANNEL(expected).key("dataInput").commit(),
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(output_schema)
            .commit(),
            VECTOR_STRING_ELEMENT(expected)
            .key("fastSources")
            .displayedName("Fast data sources")
            .description(
                "Sources to get data from. Only incoming hashes from these sources "
                "will be processed. This will typically be a single entry of the form: "
                "'[instrument]_DET_[detector]/DET/[channel]:xtdf'."
            )
            .assignmentOptional()
            .defaultValue([])
            .commit(),
            STRING_ELEMENT(expected)
            .key("frameFilter")
            .displayedName("Frame filter")
            .description(
                "The frame filter - if set - slices the input data. Frames not in the "
                "filter will be discarded before any processing happens and will not "
                "get to dataOutput or preview. Note that this filter goes by frame "
                "index rather than cell ID or pulse ID; set accordingly. Handle with "
                "care - an invalid filter can prevent all processing. The filter is "
                "specified as a string which is evaluated into numpy uint16 array. A "
                "valid filter could for eaxmple be 'np.arange(0, 352, 2)'."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
            UINT32_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .displayedName("Output buffer size limit")
            .unit(Unit.BYTE)
            .metricPrefix(MetricPrefix.GIGA)
            .description(
                "Corrected trains are written to shared memory locations. These are "
                "pre-allocated and re-used (circular buffer). This parameter "
                "determines how much memory to set aside for that buffer."
            )
            .assignmentOptional()
            .defaultValue(10)
            .commit(),
            VECTOR_STRING_ELEMENT(expected)
            .key("availableScenes")
            .setSpecialDisplayType(DT_SCENES)
            .readOnly()
            .initialValue(["overview"])
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("dataFormat")
            .displayedName("Data format (in/out)")
            .commit(),
            BOOL_ELEMENT(expected)
            .key("dataFormat.overrideInputAxisOrder")
            .displayedName("Override input axis order")
            .description(
                "The shape of the image data ndarray as received from the "
                "DataAggregator is sometimes wrong - the axes are actually in a "
                "different order than the ndarray shape suggests. If this flag is on, "
                "the shape of the ndarray will be overridden with the axis order which "
                "was expected."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
            STRING_ELEMENT(expected)
            .key("dataFormat.inputImageDtype")
            .displayedName("Input image data dtype")
            .description("The (numpy) dtype to expect for incoming image data.")
            .options("uint16,float32")
            .assignmentOptional()
            .defaultValue("uint16")
            .commit(),
            STRING_ELEMENT(expected)
            .key("dataFormat.outputImageDtype")
            .displayedName("Output image data dtype")
            .description(
                "The (numpy) dtype to use for outgoing image data. Input is cast to "
                "float32, corrections are applied, and only then will the result be "
                "cast to outputImageDtype. Be aware that casting to integer type "
                "causes truncation rather than rounding."
            )
            # TODO: consider adding rounding / binning for integer output
            .options("float16,float32,uint16")
            .assignmentOptional()
            .defaultValue("float32")
            .commit(),
            # important: determines shape of data as going into correction
            UINT32_ELEMENT(expected)
            .key("dataFormat.pixelsX")
            .displayedName("Pixels x")
            .description("Number of pixels of image data along X axis")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            UINT32_ELEMENT(expected)
            .key("dataFormat.pixelsY")
            .displayedName("Pixels y")
            .description("Number of pixels of image data along Y axis")
            .assignmentOptional()
            .defaultValue(128)
            .commit(),
            UINT32_ELEMENT(expected)
            .key("dataFormat.memoryCells")
            .displayedName("Memory cells")
            .description("Full number of memory cells in incoming data")
            .assignmentOptional()
            .noDefaultValue()  # subclass will want to set a default value
            .commit(),
            UINT32_ELEMENT(expected)
            .key("dataFormat.filteredFrames")
            .displayedName("Frames after filter")
            .description("Number of frames left after applying frame filter")
            .readOnly()
            .initialValue(0)
            .commit(),
            STRING_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .displayedName("Output axis order")
            .description(
                "Axes of main data output can be reordered after correction. Axis "
                "order is specified as string consisting of 'x', 'y', and 'c', with "
                "the latter indicating the memory cell axis. The default value of "
                "'cxy' puts pixels on the fast axes."
            )
            .assignmentOptional()
            .defaultValue("cxy")
            .commit(),
            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.inputDataShape")
            .displayedName("Input data shape")
            .description(
                "Image data shape in incoming data (from reader / DAQ). This value is "
                "computed from pixelsX, pixelsY, and memoryCells - this field just "
                "shows what is currently expected."
            )
            .readOnly()
            .initialValue([])
            .commit(),
            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.outputDataShape")
            .displayedName("Output data shape")
            .description(
                "Image data shape for data output from this device. This value is "
                "computed from pixelsX, pixelsY, and the size of the frame filter - "
                "this field just shows what is currently expected."
            )
            .readOnly()
            .initialValue([])
            .commit(),
        )

        (
            SLOT_ELEMENT(expected)
            .key("loadMostRecentConstants")
            .displayedName("Load most recent constants")
            .description(
                "Calling this slot will flush all constant buffers and cause the "
                "device to start querying CalCat for the most recent constants - all "
                "constants applicable for this device - available with the currently "
                "set constant parameters. This is typically called after "
                "instantiating pipeline, after changing parameters, or after "
                "generating new constants."
            )
            .commit()
        )

        (
            NODE_ELEMENT(expected).key("preview").displayedName("Preview").commit(),
            OUTPUT_CHANNEL(expected)
            .key("preview.outputRaw")
            .dataSchema(preview_schema)
            .commit(),
            OUTPUT_CHANNEL(expected)
            .key("preview.outputCorrected")
            .dataSchema(preview_schema)
            .commit(),
            BOOL_ELEMENT(expected)
            .key("preview.enable")
            .displayedName("Enable preview data generation")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
            INT32_ELEMENT(expected)
            .key("preview.index")
            .displayedName("Index (or stat) for preview")
            .description(
                "If this value is ≥ 0, the corresponding index (frame, cell, or pulse) "
                "will be sliced for the preview output. If this value is ≤ 0, preview "
                "will be one of the following stats: -1: max, -2: mean, -3: sum, -4: "
                "stdev. These stats are computed across memory cells."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            STRING_ELEMENT(expected)
            .key("preview.selectionMode")
            .displayedName("Index selection mode")
            .description(
                "The value of preview.index can be used in multiple ways, controlled "
                "by this value. If this is set to 'frame', preview.index is sliced "
                "directly from data. If 'cell' (or 'pulse') is selected, I will look "
                "at cell (or pulse) table for the requested cell (or pulse ID). "
                "Special (stat) index values <0 are not affected by this."
            )
            .options("frame,cell,pulse")
            .assignmentOptional()
            .defaultValue("frame")
            .reconfigurable()
            .commit(),
            UINT32_ELEMENT(expected)
            .key("preview.trainIdModulo")
            .displayedName("Preview train stride")
            .description(
                "Preview will only be generated for trains whose ID modulo this "
                "number is zero. Higher values means less frequent preview updates. "
                "Keep in mind that the GUI has limited refresh rate of 2 Hz. Should "
                "take extra care if DAQ train stride is >1."
            )
            .assignmentOptional()
            .defaultValue(6)
            .reconfigurable()
            .commit(),
        )

        # just measurements and counters to display
        (
            UINT64_ELEMENT(expected)
            .key("trainId")
            .displayedName("Train ID")
            .description("ID of latest train processed by this device.")
            .readOnly()
            .initialValue(0)
            .commit(),
            NODE_ELEMENT(expected)
            .key("performance")
            .displayedName("Performance measures")
            .commit(),
            DOUBLE_ELEMENT(expected)
            .key("performance.processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(100)
            .info("Processing not fast enough for full speed")
            .needsAcknowledging(False)
            .commit(),
            DOUBLE_ELEMENT(expected)
            .key("performance.rate")
            .displayedName("Rate")
            .description(
                "Actual rate with which this device gets, processes, and sends trains. "
                "This is a simple windowed moving average."
            )
            .unit(Unit.HERTZ)
            .readOnly()
            .initialValue(0)
            .commit(),
        )

        # this node will be filled out by subclass
        (
            NODE_ELEMENT(expected)
            .key("corrections")
            .displayedName("Correction steps")
            .commit(),
        )

    def __init__(self, config):
        self._schema_cache = {
            k: config.get(k) for k in self._schema_cache_fields if config.has(k)
        }
        super().__init__(config)
        self.updateState(State.INIT)

        if not sorted(config.get("dataFormat.outputAxisOrder")) == ["c", "x", "y"]:
            # TODO: figure out how to get this information to operator
            self.log_status_error("Invalid output axis order string")
            return

        self.KARABO_ON_INPUT("dataInput", self.input_handler)
        self.KARABO_ON_EOS("dataInput", self.handle_eos)

        self.sources = set(config.get("fastSources"))

        self.input_data_dtype = np.dtype(config.get("dataFormat.inputImageDtype"))
        self.output_data_dtype = np.dtype(config.get("dataFormat.outputImageDtype"))
        self.kernel_runner = None  # must call _update_buffers to initialize
        self.registerInitialFunction(self._update_frame_filter)
        self.registerInitialFunction(self._update_buffers)

        self.calcat_friend = self._calcat_friend_class(
            self, pathlib.Path.cwd() / "calibration-client-secrets.json"
        )

        self._correction_flag_enabled = self._correction_flag_class.NONE
        self._correction_flag_preview = self._correction_flag_class.NONE

        self._shmem_buffer = None
        self._processing_time_ema = utils.ExponentialMovingAverage(alpha=0.3)
        self._rate_tracker = utils.WindowRateTracker()

        self._buffered_status_update = Hash(
            "trainId",
            0,
            "performance.rate",
            0,
            "performance.processingTime",
            0,
        )
        self._last_processing_started = 0  # used for processing time and timeout
        self._rate_update_timer = utils.RepeatingTimer(
            interval=1,
            callback=self._update_rate_and_state,
        )
        self._buffer_lock = threading.Lock()
        self.KARABO_SLOT(self.loadMostRecentConstants)
        self.KARABO_SLOT(self.requestScene)

        if parse_version(karaboVersion) >= parse_version("2.11"):
            # TODO: the CalCatFriend could add these for us
            # note: overly complicated for closure to work
            def make_wrapper_capturing_constant(constant):
                def aux():
                    self.calcat_friend.get_specific_constant_version_and_call_me_back(
                        constant, self._load_constant_to_runner
                    )

                return aux

            for constant in self._constant_enum_class:
                slot_name = f"foundConstants.{constant.name}.overrideConstantVersion"
                meth_name = slot_name.replace(".", "_")
                self.KARABO_SLOT(
                    make_wrapper_capturing_constant(constant),
                    slotName=meth_name,
                )

    def __del__(self):
        del self._shmem_buffer

    def preReconfigure(self, config):
        for ts_path in (
            "constantParameters.deviceMappingSnapshotAt",
            "constantParameters.constantVersionEventAt",
        ):
            if config.has(ts_path):
                ts_string = config.get(ts_path)
                if ts_string.strip() == "":
                    config.set(ts_path, "")
                else:
                    try:
                        timestamp = dateutil.parser.isoparse(ts_string)
                    except ValueError as error:
                        self.log_status_warn(f"Failed to parse {ts_path}; {error}")
                        config.erase(ts_path)
                    else:
                        config.set(ts_path, timestamp.isoformat())
        if config.has("constantParameters.deviceMappingSnapshotAt"):
            self.calcat_friend.flush_pdu_mapping()
        self._prereconfigure_update_hash = config

    def postReconfigure(self):
        if not hasattr(self, "_prereconfigure_update_hash"):
            self.log_status_warn("postReconfigure without knowing update hash")
            return

        update = self._prereconfigure_update_hash
        for path in update.getPaths():
            if path in self._schema_cache_fields:
                self._schema_cache[path] = update.get(path)

        if update.has("frameFilter"):
            with self._buffer_lock:
                self._update_frame_filter()
        if any(
            update.has(shape_param)
            for shape_param in (
                "dataFormat.pixelsX",
                "dataFormat.pixelsY",
                "dataFormat.memoryCells",
                "constantParameters.memoryCells",
                "frameFilter",
            )
        ):
            with self._buffer_lock:
                self._update_buffers()
        # TODO: only call this if they are changed (is cheap, though)
        self._update_correction_flags()

    def loadMostRecentConstants(self):
        self.flush_constants()
        self.calcat_friend.flush_constants()
        for constant in self._constant_enum_class:
            self.calcat_friend.get_constant_version_and_call_me_back(
                constant, self._load_constant_to_runner
            )

    def flush_constants(self):
        """Reset constant buffers and disable corresponding correction steps"""
        for correction_step, _ in self._correction_field_names:
            self.set(f"corrections.{correction_step}.available", False)
        self.kernel_runner.flush_buffers()
        self._update_correction_flags()

    def log_status_info(self, msg):
        self.log.INFO(msg)
        self.set("status", msg)

    def log_status_warn(self, msg):
        self.log.WARN(msg)
        self.set("status", msg)

    def log_status_error(self, msg):
        self.set("status", msg)
        self.log.ERROR(msg)
        self.updateState(State.ERROR)

    def set(self, *args):
        """Wrapper around PythonDevice.set to enable caching "hot" schema elements"""
        # TODO: handle other cases of PythonDevice.set arguments
        if len(args) == 2 and not isinstance(args[0], Hash):
            key, value = args
            if key in self._schema_cache_fields:
                self._schema_cache[key] = value
        super().set(*args)

    def requestScene(self, params):
        payload = Hash()
        name = params.get("name", default="")
        payload["name"] = name
        payload["success"] = True
        if name == "overview":
            payload["data"] = scenes.correction_device_overview_scene(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
            )
        elif name.startswith("browse_schema"):
            if ":" in name:
                prefix = name[len("browse_schema:") :]
            else:
                prefix = "managed"
            payload["data"] = scenes.recursive_subschema_scene(
                self.getInstanceId(),
                self.getFullSchema(),
                prefix,
            )
        else:
            payload["success"] = False
        response = Hash()
        response["type"] = "deviceScene"
        response["origin"] = self.getInstanceId()
        response["payload"] = payload
        self.reply(response)

    def _write_output(self, data, old_metadata):
        """For dataOutput: reusing incoming data hash and setting source and timestamp
        to be same as input"""
        metadata = ChannelMetaData(
            old_metadata.get("source"),
            Timestamp.fromHashAttributes(old_metadata.getAttributes("timestamp")),
        )

        channel = self.signalSlotable.getOutputChannel("dataOutput")
        channel.write(data, metadata, False)
        channel.update()

    def _write_combiner_previews(self, channel_data_pairs, train_id, source):
        # TODO: send as ImageData (requires updated assembler)
        # TODO: allow sending *all* frames for commissioning (request: Jola)
        preview_hash = Hash()
        preview_hash.set("image.trainId", train_id)

        # note: have to construct because setting .tid after init is broken
        timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        metadata = ChannelMetaData(source, timestamp)
        for channel_name, data in channel_data_pairs:
            preview_hash.set("image.data", data)
            channel = self.signalSlotable.getOutputChannel(channel_name)
            channel.write(preview_hash, metadata, False)
            channel.update()

    def _update_correction_flags(self):
        """Based on constants loaded and settings, update bit mask flags for kernel"""
        available = self._correction_flag_class.NONE
        enabled = self._correction_flag_class.NONE
        preview = self._correction_flag_class.NONE
        for field_name, flag in self._correction_field_names:
            if self.get(f"corrections.{field_name}.available"):
                available |= flag
            if self.get(f"corrections.{field_name}.enable"):
                enabled |= flag
            if self.get(f"corrections.{field_name}.preview"):
                preview |= flag
        enabled &= available
        preview &= available
        self._correction_flag_enabled = enabled
        self._correction_flag_preview = preview
        self.log.DEBUG(f"Corrections for dataOutput: {str(enabled)}")
        self.log.DEBUG(f"Corrections for preview: {str(preview)}")

    def _update_frame_filter(self):
        """Parse frameFilter string (if set) and update cached filter array. Will set
        dataFormat.filteredFrames, so one will typically want to call _update_buffers
        afterwards."""
        filter_string = self.get("frameFilter")
        if filter_string.strip() == "":
            self._frame_filter = None
            self.set("dataFormat.filteredFrames", self.get("dataFormat.memoryCells"))
        else:
            self._frame_filter = np.array(eval(filter_string), dtype=np.uint16)
            self.set("dataFormat.filteredFrames", self._frame_filter.size)
            if self._frame_filter.min() < 0 or self._frame_filter.max() >= self.get(
                "dataFormat.memoryCells"
            ):
                self.log_status_warn("Invalid frame filter set, expect exceptions!")

    def _update_buffers(self):
        """(Re)initialize buffers / kernel runner according to expected data shapes"""
        self.log.INFO("Updating buffers according to data shapes")
        # reflect the axis reordering in the expected output shape
        self.set("dataFormat.inputDataShape", list(self.input_data_shape))
        self.set("dataFormat.outputDataShape", list(self.output_data_shape))
        self.log.INFO(f"Input shape: {self.input_data_shape}")
        self.log.INFO(f"Output shape: {self.output_data_shape}")

        if self._shmem_buffer is None:
            shmem_buffer_name = self.getInstanceId() + ":dataOutput"
            memory_budget = self.get("outputShmemBufferSize") * 2 ** 30
            self.log.INFO(f"Opening new shmem buffer: {shmem_buffer_name}")
            self._shmem_buffer = shmem_utils.ShmemCircularBuffer(
                memory_budget,
                self.output_data_shape,
                self.output_data_dtype,
                shmem_buffer_name,
            )
            self.log.INFO("Trying to pin the shmem buffer memory")
            self._shmem_buffer.cuda_pin()
        else:
            self._shmem_buffer.change_shape(self.output_data_shape)

        self.kernel_runner = self._kernel_runner_class(
            self.get("dataFormat.pixelsX"),
            self.get("dataFormat.pixelsY"),
            self.get("dataFormat.filteredFrames"),
            int(self.get("constantParameters.memoryCells")),
            input_data_dtype=self.input_data_dtype,
            output_data_dtype=self.output_data_dtype,
            **self._kernel_runner_init_args,
        )

        # TODO: gracefully handle change in constantParameters.memoryCells
        with self.calcat_friend.cached_constants_lock:
            for (
                constant,
                data,
            ) in self.calcat_friend.cached_constants.items():
                self.log_status_info(f"Reload constant {constant}")
                self._load_constant_to_runner(constant, data)

    def input_handler(self, input_channel):
        """Main handler for data input: Do a few simple checks to determine whether to
        even try processing. If yes, will pass data and information to process_data
        method provided by subclass."""

        # Is device even ready for this?
        state = self._schema_cache["state"]
        if state is State.ERROR:
            # in this case, we should have already issued warning
            return
        elif self.kernel_runner is None:
            self.log_status_warn("Received data, but have not initialized kernels yet")
            return

        all_metadata = input_channel.getMetaData()
        for input_index in range(input_channel.size()):
            self._last_processing_started = default_timer()
            data_hash = input_channel.read(input_index)
            metadata = all_metadata[input_index]
            source = metadata.get("source")

            if source not in self.sources:
                self.log_status_info(f"Ignoring hash with unknown source {source}")
                return
            elif not data_hash.has("image"):
                self.log_status_info("Ignoring hash without image node")
                return

            train_id = metadata.getAttribute("timestamp", "tid")
            cell_table = np.squeeze(data_hash.get("image.cellId"))
            if len(cell_table.shape) == 0:
                self.log_status_warn(
                    "cellId had 0 dimensions. DAQ may not be sending data."
                )
                return

            # no more common reasons to skip input, so go to processing
            if state is State.ON:
                self.updateState(State.PROCESSING)
                self.log_status_info("Processing data")

            correction_cell_num = self._schema_cache["constantParameters.memoryCells"]
            cell_table_max = np.max(cell_table)
            if cell_table_max >= correction_cell_num:
                self.log_status_info(
                    f"Max cell ID ({cell_table_max}) exceeds range for loaded "
                    f"constants ({correction_cell_num} cells). Some frames will not be "
                    "corrected."
                )

            image_data = data_hash.get("image.data")
            if image_data.shape[0] != self._schema_cache["dataFormat.memoryCells"]:
                self.log_status_info(
                    f"Updating new input shape {image_data.shape}, updating buffers"
                )
                self.set("dataFormat.memoryCells", image_data.shape[0])
                with self._buffer_lock:
                    self._update_frame_filter()
                    self._update_buffers()

            # DataAggregator typically tells us the wrong axis order
            if self._schema_cache["dataFormat.overrideInputAxisOrder"]:
                expected_shape = self.input_data_shape
                if expected_shape != image_data.shape:
                    image_data.shape = expected_shape

            do_generate_preview = (
                train_id % self._schema_cache["preview.trainIdModulo"] == 0
                and self._schema_cache["preview.enable"]
            )

            with self._buffer_lock:
                self.process_data(
                    data_hash,
                    metadata,
                    source,
                    train_id,
                    image_data,
                    cell_table,
                    do_generate_preview,
                )
            self._buffered_status_update.set("trainId", train_id)
            self._processing_time_ema.update(
                default_timer() - self._last_processing_started
            )
            self._rate_tracker.update()

    def _update_rate_and_state(self):
        self._buffered_status_update.set("performance.rate", self._rate_tracker.get())
        self._buffered_status_update.set(
            "performance.processingTime", self._processing_time_ema.get() * 1000
        )
        # trainId in _buffered_status_update should be updated in input handler

        self.set(self._buffered_status_update)

        if default_timer() - self._last_processing_started > PROCESSING_STATE_TIMEOUT:
            if self.get("state") is State.PROCESSING:
                self.updateState(State.ON)
                self.log_status_info(
                    f"No new train in {PROCESSING_STATE_TIMEOUT} s, switching state."
                )

    def handle_eos(self, channel):
        self.updateState(State.ON)
        self.signalEndOfStream("dataOutput")


def add_correction_step_schema(schema, managed_keys, field_flag_mapping):
    """Using the fields in the provided mapping, will add nodes to schema

    field_flag_mapping is assumed to be iterable of pairs where first entry in each
    pair is the name of a correction step as it will appear in device schema (second
    entry - typically an enum field - is ignored). For correction step, a node and some
    booleans are added to the schema and the toggleable booleans are added to
    managed_keys. Subclass can customize / add additional keys under node later.

    This method should be called in expectedParameters of subclass after the same for
    BaseCorrection has been called. Would be nice to include in BaseCorrection instead,
    but that is tricky: static method of superclass will need _correction_field_names
    of subclass or device server gets mad. A nice solution with classmethods would be
    welcome.
    """

    for field_name, _ in field_flag_mapping:
        node_name = f"corrections.{field_name}"
        (
            NODE_ELEMENT(schema).key(node_name).commit(),
            BOOL_ELEMENT(schema)
            .key(f"{node_name}.available")
            .displayedName("Available")
            .description(
                "This boolean indicates whether the necessary constants have been "
                "loaded for this correction step to be applied. Enabling the "
                "correction will have no effect unless this is True."
            )
            .readOnly()
            .initialValue(False)
            .commit(),
            BOOL_ELEMENT(schema)
            .key(f"{node_name}.enable")
            .displayedName("Enable")
            .description(
                "Controls whether to apply this correction step for main data "
                "output - subject to availability."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
            BOOL_ELEMENT(schema)
            .key(f"{node_name}.preview")
            .displayedName("Preview")
            .description(
                "Whether to apply this correction step for corrected preview "
                "output - subject to availability."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add(f"{node_name}.enable")
        managed_keys.add(f"{node_name}.preview")
