import pathlib

import cupy
import jinja2
import numpy as np

from . import utils


class BaseGpuRunner:
    """Class to handle GPU buffers and execution of CUDA kernels on image data

    All GPU buffers are kept within this class and it is intentionally very stateful.
    This generally means that you will want to load data into it and then do something.
    Typical usage in correct order:

    1. instantiate
    2. load constants
    3. load_data
    4. load_cell_table
    5. correct
    6a. reshape (only here does data transfer back to host)
    6b. compute_preview (optional)

    repeat from 2. or 3.

    In case no constants are available / correction is not desired, can skip 3 and 4 and
    pass CorrectionFlags.NONE to correct(...). Generally, user must handle which
    correction steps are appropriate given the constants loaded so far.
    """

    # These must be set by subclass
    _kernel_source_filename = None
    _corrected_axis_order = None

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        _src_dir = pathlib.Path(__file__).absolute().parent
        # subclass must define _kernel_source_filename
        with (_src_dir / self._kernel_source_filename).open("r") as fd:
            self._kernel_template = jinja2.Template(fd.read())

        self.pixels_x = pixels_x
        self.pixels_y = pixels_y
        self.memory_cells = memory_cells
        if constant_memory_cells == 0:
            # if not set, guess same as input; may save one recompilation
            self.constant_memory_cells = memory_cells
        else:
            self.constant_memory_cells = constant_memory_cells
        # preview will only be single memory cell
        self.preview_shape = (self.pixels_x, self.pixels_y)
        self.input_data_dtype = input_data_dtype
        self.output_data_dtype = output_data_dtype

        self._init_kernels()

        # reuse buffers for input / output
        self.cell_table_gpu = cupy.empty(self.memory_cells, dtype=np.uint16)
        self.input_data_gpu = cupy.empty(self.input_shape, dtype=input_data_dtype)
        self.processed_data_gpu = cupy.empty(
            self.processed_shape, dtype=output_data_dtype
        )
        self.reshaped_data_gpu = None  # currently not reusing buffer

        # default preview layers: raw and corrected (subclass can extend)
        self.preview_buffer_getters = [
            self._get_raw_for_preview,
            self._get_corrected_for_preview,
        ]

    # to get data from respective buffers to cell, x, y shape for preview computation
    def _get_raw_for_preview(self):
        """Should return view of self.input_data_gpu with shape (cell, x/y, x/y)"""
        raise NotImplementedError()

    def _get_corrected_for_preview(self):
        """Should return view of self.processed_data_gpu with shape (cell, x/y, x/y)"""
        raise NotImplementedError()

    def flush_buffers(self):
        """Optional reset GPU buffers (implement in subclasses which need this)"""
        pass

    def correct(self, flags):
        """Correct (already loaded) image data according to flags

        Subclass must define this method. It should assume that image data, cell table,
        and other data (including constants) has already been loaded. It should
        probably run some GPU kernel and output should go into self.processed_data_gpu.

        Keep in mind that user only gets output from compute_preview or reshape
        (either of these should come after correct).

        The submodules providing subclasses should have some IntFlag enums defining
        which flags are available to pass along to the kernel. A zero flag should allow
        the kernel to do no actual correction - but still copy the data between buffers
        and cast it to desired output type.

        """
        raise NotImplementedError()

    def reshape(self, output_order, out=None):
        """Move axes to desired output order and copy to host memory

        The out parameter is passed directly to the get function of GPU array: if
        None, then a new ndarray (in host memory) is returned. If not None, then data
        will be loaded into the provided array, which must match shape / dtype.
        """
        # TODO: avoid copy
        if output_order == self._corrected_axis_order:
            self.reshaped_data_gpu = self.processed_data_gpu
        else:
            self.reshaped_data_gpu = cupy.transpose(
                self.processed_data_gpu,
                utils.transpose_order(self._corrected_axis_order, output_order),
            )

        return self.reshaped_data_gpu.get(out=out)

    def load_data(self, raw_data):
        self.input_data_gpu.set(np.squeeze(raw_data))

    def load_cell_table(self, cell_table):
        self.cell_table_gpu.set(cell_table)

    def compute_previews(self, preview_index):
        """Generate single slice or reduction preview of raw and corrected data

        Special values of preview_index are -1 for max, -2 for mean, -3 for sum, and
        -4 for stdev (across cells).

        Note that preview_index is taken from data without checking cell table.
        Caller has to figure out which index along memory cell dimension they
        actually want to preview in case it needs to be a specific pulse.

        Will reuse data from corrected output buffer. Therefore, correct(...) must have
        been called with the appropriate flags before compute_preview(...).
        """

        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.memory_cells:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        # TODO: enum around reduction type
        return tuple(
            self._compute_a_preview(image_data=getter(), preview_index=preview_index)
            for getter in self.preview_buffer_getters
        )

    def _compute_a_preview(self, image_data, preview_index):
        """image_data must have cells on first axis; X and Y order is not important
        here for now (and can differ between AGIPD and DSSC)"""
        if preview_index >= 0:
            # TODO: reuse pinned buffers for this
            return image_data[preview_index].astype(np.float32).get()
        elif preview_index == -1:
            # TODO: confirm that max is pixel and not integrated intensity
            # separate from next case because dtype not applicable here
            return cupy.nanmax(image_data, axis=0).astype(cupy.float32).get()
        elif preview_index in (-2, -3, -4):
            stat_fun = {
                -1: cupy.nanmax,
                -2: cupy.nanmean,
                -3: cupy.nansum,
                -4: cupy.nanstd,
            }[preview_index]
            return stat_fun(image_data, axis=0, dtype=cupy.float32).get()

    def update_block_size(self, full_block, target_shape=None):
        """Compute grid such that thread block grid covers target shape

        Execution is scheduled with 3d "blocks" of CUDA threads, tuning can affect
        performance. Correction kernels are "monolithic" for simplicity (i.e. each
        logical thread handles one entry in output data), so in each dimension we
        parallelize, grid * block >= length.

        Note that individual kernels must themselves check whether they go out of
        bounds; grid dimensions get rounded up in case ndarray size is not multiple of
        block size.

        """
        if target_shape is None:
            target_shape = self.processed_shape
        assert len(full_block) == 3
        self.full_block = tuple(full_block)
        self.full_grid = tuple(
            utils.ceil_div(a_length, block_length)
            for (a_length, block_length) in zip(target_shape, full_block)
        )
