import copy
import enum
import functools
import json
import pathlib
import threading
import typing

import calibration_client
import h5py
import numpy as np
from calibration_client.modules import (
    Calibration,
    CalibrationConstant,
    CalibrationConstantVersion,
    Detector,
    DetectorType,
    PhysicalDetectorUnit,
)
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    NODE_ELEMENT,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
)
from karabo import version as karaboVersion
from pkg_resources import parse_version

from . import utils


class ConditionNotFound(Exception):
    pass


class DetectorNotFound(Exception):
    pass


class ModuleNotFound(Exception):
    pass


class CalibrationNotFound(Exception):
    pass


class CalibrationClientConfigError(Exception):
    pass


def _add_status_schema_from_enum(schema, prefix, enum_class):
    for constant in enum_class:
        constant_node = f"{prefix}.{constant.name}"
        (
            NODE_ELEMENT(schema).key(constant_node).commit(),
            BOOL_ELEMENT(schema)
            .key(f"{constant_node}.found")
            .readOnly()
            .initialValue(False)
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.validFrom")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.calibrationId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.conditionId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.constantId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{constant_node}.constantVersionId")
            .description(
                "This field is editable - if for any reason a specific constant "
                "version is desired, the constant version ID (as used in CalCat) can "
                "be set here and the slot below can be called to load this particular "
                "version, overriding the automatic loading of latest constants."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
        )
        if parse_version(karaboVersion) >= parse_version("2.11"):
            (
                SLOT_ELEMENT(schema)
                .key(f"{constant_node}.overrideConstantVersion")
                .displayedName("Override constant version")
                .commit(),
            )


class OperatingConditions(dict):
    # TODO: support deviation?
    def encode(self):
        return {
            "parameters": [
                {
                    "parameter_name": key,
                    "lower_deviation_value": 0.0,
                    "upper_deviation_value": 0.0,
                    "flg_available": False,
                    "value": value,
                }
                for (key, value) in self.items()
            ]
        }

    def __hash__(self):
        # this takes me back to pre-screening interview time...
        return hash(tuple(sorted(self.items())))


class BaseCalcatFriend:
    """Base class for CalCat friends - handles interacting with CalCat for the device

    A CalCat friend uses the device schema to build up parameters for CalCat queries.
    It focuses on two nodes (added by static method add_schema): param_prefix and
    status_prefix. The former is primarily used to get parameters which are (via
    condition methods - see for example dark_condition of DsscCalcatFriend) used
    to look for constants. The latter is primarily used to give user information
    about what was found.
    """

    _constant_enum_class = None  # subclass should set
    _constants_need_conditions = None  # subclass should set

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
        detector_type,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        """Add elements needed by this object to device's schema (expectedSchema)

        All elements added to schema go under prefixes which should end with name of
        node which does not exist yet. To change default values and add more fields,
        extend this method in subclass.

        The param_prefix node will hold all the parameters needed to build constant
        condition dicts for querying CalCat. These values are set either directly on
        the device or via manager and this class gets them from the device using helper
        function _get_param. See for example AgipdCalcatFriend.dark_condition.

        The status_prefix node is used to report information about what was found in
        CalCat. This class will update the values on the device using the helper
        function _set_status. This should not need to happen in subclass methods.
        """

        (
            NODE_ELEMENT(schema)
            .key(param_prefix)
            .displayedName("Constant retrieval parameters")
            .commit(),
            NODE_ELEMENT(schema)
            .key(status_prefix)
            .displayedName("Constants retrieved")
            .commit(),
        )

        # Parameters which any detector would probably have (extend this in subclass)
        # TODO: probably switch to floating point for everything, including mem cells
        (
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.deviceMappingSnapshotAt")
            .displayedName("Snapshot timestamp (for device mapping)")
            .description(
                "CalCat supports querying with a specific snapshot of the database. "
                "When playing back a run from the file system, this feature is useful "
                "to look up the device mapping at the time of the run. If this field "
                "is left empty, the latest device mapping is used. Date format should "
                "be 'YYYY-MM-DD' with optional time of day starting with 'T' followed "
                "by 'hh:mm:ss.mil+02:00'."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.constantVersionEventAt")
            .displayedName("Event at timestamp (for constant version)")
            .description("TODO")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.detectorType")
            .displayedName("Detector type name")
            .description(
                "Name of detector type in CalCat; typically has suffix '-Type'"
            )
            .readOnly()
            .initialValue(detector_type)
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.detectorTypeId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.detectorName")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.detectorId")
            .readOnly()
            .initialValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.karaboDa")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
            STRING_ELEMENT(schema)
            .key(f"{param_prefix}.moduleId")
            .readOnly()
            .initialValue("")
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{param_prefix}.memoryCells")
            .assignmentOptional()
            .noDefaultValue()
            .reconfigurable()
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsX")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),
            UINT32_ELEMENT(schema)
            .key(f"{param_prefix}.pixelsY")
            .assignmentOptional()
            .defaultValue(128)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.biasVoltage")
            .assignmentOptional()
            .defaultValue(300)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add(f"{param_prefix}.deviceMappingSnapshotAt")
        managed_keys.add(f"{param_prefix}.constantVersionEventAt")
        managed_keys.add(f"{param_prefix}.memoryCells")
        managed_keys.add(f"{param_prefix}.pixelsX")
        managed_keys.add(f"{param_prefix}.pixelsY")
        managed_keys.add(f"{param_prefix}.biasVoltage")

    def __init__(
        self,
        device,
        secrets_fn: pathlib.Path,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        self.device = device
        self.param_prefix = param_prefix
        self.status_prefix = status_prefix
        self.cached_constants = {}
        self.cached_constants_lock = threading.Lock()
        # api lock used to force queries to be sequential (SSL issue on ONC)
        self.api_lock = threading.Lock()

        if not secrets_fn.is_file():
            self.device.log_status_warn(
                f"Missing CalCat secrets file (expected {secrets_fn})"
            )
        with secrets_fn.open("r") as fd:
            calcat_secrets = json.load(fd)

        self.caldb_store = pathlib.Path(calcat_secrets["caldb_store_path"])
        if not self.caldb_store.is_dir():
            raise ValueError(f"caldb_store location '{self.caldb_store}' is not dir")

        self.device.log.INFO(f"Connecting to CalCat at {calcat_secrets['base_url']}")
        base_url = calcat_secrets["base_url"]
        self.client = calibration_client.CalibrationClient(
            client_id=calcat_secrets["client_id"],
            client_secret=calcat_secrets["client_secret"],
            user_email=calcat_secrets["user_email"],
            base_api_url=f"{base_url}/api/",
            token_url=f"{base_url}/oauth/token",
            refresh_url=f"{base_url}/oauth/token",
            auth_url=f"{base_url}/oauth/authorize",
            scope="public",
            session_token=None,
        )
        self.device.log_status_info("CalCat connection established")

    def _get_param(self, key):
        """Helper to get value from attached device schema"""
        return self.device.get(f"{self.param_prefix}.{key}")

    def _set_param(self, key, value):
        self.device.set(f"{self.param_prefix}.{key}", value)

    def _get_status(self, constant, key):
        return self.device.get(f"{self.status_prefix}.{constant.name}.{key}")

    def _set_status(self, constant, key, value):
        """Helper to update information about found constants on device"""
        self.device.set(f"{self.status_prefix}.{constant.name}.{key}", value)

    # Python 3.6 does not have functools.cached_property or even functools.cache
    @property
    @functools.lru_cache()
    def detector_id(self):
        detector_name = self._get_param("detectorName")
        resp = Detector.get_by_identifier(self.client, detector_name)
        self._check_resp(resp, DetectorNotFound, f"Detector {detector_name} not found")
        res = resp["data"]["id"]
        self._set_param("detectorId", str(res))
        return res

    @property
    @functools.lru_cache()
    def detector_type_id(self):
        detector_type = self._get_param("detectorType")
        resp = DetectorType.get_by_name(self.client, detector_type)
        self._check_resp(
            resp, DetectorNotFound, f"Detector type {detector_type} not found"
        )
        res = resp["data"]["id"]
        self._set_param("detectorTypeId", str(res))
        return res

    @property
    @functools.lru_cache()
    def pdus(self):
        resp = PhysicalDetectorUnit.get_all_by_detector(
            self.client, self.detector_id, self._get_param("deviceMappingSnapshotAt")
        )
        self._check_resp(resp, warning="Failed to retrieve module mapping")
        for irrelevant_key in ("detector", "detector_type", "flg_available"):
            for pdu in resp["data"]:
                del pdu[irrelevant_key]
        return resp["data"]

    @property
    @functools.lru_cache()
    def _karabo_da_to_float_uuid(self):
        return {pdu["karabo_da"]: pdu["float_uuid"] for pdu in self.pdus}

    @property
    @functools.lru_cache()
    def _karabo_da_to_id(self):
        return {pdu["karabo_da"]: pdu["id"] for pdu in self.pdus}

    def flush_pdu_mapping(self):
        for attr in ("pdus", "_karabo_da_to_float_uuid", "_karabo_da_to_id"):
            if hasattr(self, attr):
                delattr(self, attr)
        self._set_param("moduleId", "")

    @utils.threadsafe_cache
    def calibration_id(self, calibration_name: str):
        resp = Calibration.get_by_name(self.client, calibration_name)
        self._check_resp(
            resp, CalibrationNotFound, f"Calibration type {calibration_name} not found!"
        )
        return resp["data"]["id"]

    @utils.threadsafe_cache
    def condition_id(self, pdu, condition):
        # modifying condition parameter messes with cache
        condition_with_detector = copy.copy(condition)
        condition_with_detector["Detector UUID"] = pdu
        self.device.log.DEBUG(f"Look for condition: {condition_with_detector}")
        resp = self.client.search_possible_conditions_from_dict(
            "", condition_with_detector.encode()
        )
        self._check_resp(
            resp,
            ConditionNotFound,
            f"Failed to find condition {condition} for pdu {pdu}",
        )
        return resp["data"][0]["id"]

    @utils.threadsafe_cache
    def constant_id(self, calibration_id, condition_id):
        resp = CalibrationConstant.get_by_uk(
            self.client,
            calibration_id=calibration_id,
            detector_type_id=self.detector_type_id,
            condition_id=condition_id,
        )
        self._check_resp(resp, warning="Failed to retrieve constant ID")
        return resp["data"]["id"]

    def get_constant_version(self, constant):
        # TODO: catch exceptions, give warnings appropriately
        karabo_da = self._get_param("karaboDa")
        self.device.log_status_info(f"Attempting to find {constant} for {karabo_da}")

        if karabo_da not in self._karabo_da_to_float_uuid:
            self.device.log_status_warn(
                f"Module {karabo_da} not found in mapping, check configuration!"
            )
            raise ModuleNotFound(f"Module map did not include {karabo_da}")
        self._set_param("moduleId", str(self._karabo_da_to_id[karabo_da]))

        if isinstance(constant, str):
            constant = self._constant_enum_class[constant]

        calibration_id = self.calibration_id(constant.name)
        self._set_status(constant, "calibrationId", calibration_id)

        condition = self._constants_need_conditions[constant]()
        condition_id = self.condition_id(
            self._karabo_da_to_float_uuid[karabo_da], condition
        )
        self._set_status(constant, "conditionId", condition_id)

        constant_id = self.constant_id(
            calibration_id=calibration_id, condition_id=condition_id
        )
        self._set_status(constant, "constantId", constant_id)

        resp = CalibrationConstantVersion.get_closest_by_time(
            self.client,
            calibration_constant_ids=[constant_id],
            physical_detector_unit_id=self._karabo_da_to_id[karabo_da],
            event_at=self._get_param("constantVersionEventAt"),
            snapshot_at=None,
        )
        self._check_resp(resp, warning="Failed to find calibration constant version")
        # TODO: replace with start date and end date
        timestamp = resp["data"]["begin_validity_at"]
        self._set_status(constant, "validFrom", timestamp)
        self._set_status(constant, "constantVersionId", resp["data"]["id"])

        file_path = (
            self.caldb_store / resp["data"]["path_to_file"] / resp["data"]["file_name"]
        )
        # TODO: handle FileNotFoundError if we are led astray
        with h5py.File(file_path, "r") as fd:
            constant_data = np.array(fd[resp["data"]["data_set_name"]]["data"])
        with self.cached_constants_lock:
            self.cached_constants[constant] = constant_data
        self._set_status(constant, "found", True)
        self.device.log_status_info(f"Done finding {constant} for {karabo_da}")

        return constant_data

    def get_specific_constant_version(self, constant):
        # TODO: warn if PDU or constant type does not match
        # TODO: warn if result is list (happens for empty version ID)
        constant_version_id = self.device.get(
            f"{self.status_prefix}.{constant.name}.constantVersionId"
        )

        resp = CalibrationConstantVersion.get_by_id(self.client, constant_version_id)
        self._check_resp(resp, warning="Failed to find calibration constant version")
        file_path = (
            self.caldb_store / resp["data"]["path_to_file"] / resp["data"]["file_name"]
        )
        with h5py.File(file_path, "r") as fd:
            constant_data = np.array(fd[resp["data"]["data_set_name"]]["data"])
        with self.cached_constants_lock:
            self.cached_constants[constant] = constant_data
        self._set_status(constant, "validFrom", resp["data"]["begin_at"])
        self._set_status(constant, "calibrationId", "manual override")
        self._set_status(constant, "conditionId", "manual override")
        self._set_status(constant, "constantId", "manual override")
        self._set_status(constant, "constantVersionId", constant_version_id)
        self._set_status(constant, "found", True)
        return constant_data

    def get_constant_version_and_call_me_back(self, constant, callback):
        """Runs get_constant_version in thread, will call callback on completion"""
        # TODO: do we want to use asyncio / "modern" async?
        # TODO: consider moving out of this class, closer to correction device
        def aux():
            with self.api_lock:
                data = self.get_constant_version(constant)
            callback(constant, data)

        thread = threading.Thread(target=aux)
        thread.start()
        return thread

    def get_specific_constant_version_and_call_me_back(self, constant, callback):
        """Blindly load whatever CalCat points to for CCV - user must be confident that
        this CCV corresponds to correct kind of constant."""

        # TODO: warn user about all the things that go wrong
        def aux():
            with self.api_lock:
                data = self.get_specific_constant_version(constant)
            callback(constant, data)

        thread = threading.Thread(target=aux)
        thread.start()
        return thread

    def flush_constants(self):
        for constant in self._constant_enum_class:
            self._set_status(constant, "validFrom", "")
            self._set_status(constant, "found", False)

    def _check_resp(self, resp, exception=Exception, warning=None):
        # TODO: probably verify using "info" that exception is the right one
        to_raise = None
        if not resp["success"]:
            # TODO: probably more types of app_info errors?
            if resp["app_info"]:
                if "not found" in resp["info"]:
                    # this was likely the exception exception
                    to_raise = exception(resp["info"])
                else:
                    # but could also be authorization or similar issue
                    to_raise = CalibrationClientConfigError(resp["app_info"])
            to_raise = exception(resp["info"])
        if to_raise is not None:
            if warning is not None:
                self.device.log_status_warn(warning)
                raise to_raise


class AgipdConstants(enum.Enum):
    SlopesFF = enum.auto()
    ThresholdsDark = enum.auto()
    Offset = enum.auto()
    SlopesPC = enum.auto()
    BadPixelsDark = enum.auto()
    BadPixelsPC = enum.auto()
    BadPixelsFF = enum.auto()


class AgipdCalcatFriend(BaseCalcatFriend):
    _constant_enum_class = AgipdConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            AgipdConstants.ThresholdsDark: self.dark_condition,
            AgipdConstants.Offset: self.dark_condition,
            AgipdConstants.SlopesPC: self.dark_condition,
            AgipdConstants.SlopesFF: self.illuminated_condition,
            AgipdConstants.BadPixelsDark: self.dark_condition,
            AgipdConstants.BadPixelsPC: self.dark_condition,
            AgipdConstants.BadPixelsFF: self.illuminated_condition,
        }

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        super(AgipdCalcatFriend, AgipdCalcatFriend).add_schema(
            schema, managed_keys, "AGIPD-Type", param_prefix, status_prefix
        )
        schema.setDefaultValue(f"{param_prefix}.memoryCells", 352)

        (
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.acquisitionRate")
            .assignmentOptional()
            .defaultValue(1.1)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.gainSetting")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.photonEnergy")
            .assignmentOptional()
            .defaultValue(9.2)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.gainMode")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.integrationTime")
            .assignmentOptional()
            .defaultValue(12)
            .reconfigurable()
            .commit(),
        )
        managed_keys.add(f"{param_prefix}.acquisitionRate")
        managed_keys.add(f"{param_prefix}.gainSetting")
        managed_keys.add(f"{param_prefix}.photonEnergy")
        managed_keys.add(f"{param_prefix}.integrationTime")

        _add_status_schema_from_enum(schema, status_prefix, AgipdConstants)

    def dark_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Acquisition rate"] = self._get_param("acquisitionRate")
        # TODO: make configurable whether or not to include gain setting?
        integration_time = self._get_param("integrationTime")
        if integration_time != 12:
            res["Integration Time"] = integration_time
        res["Gain Setting"] = self._get_param("gainSetting")
        return res

    def illuminated_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        res["Source Energy"] = self._get_param("photonEnergy")
        res["Acquisition rate"] = self._get_param("acquisitionRate")
        integration_time = self._get_param("integrationTime")
        if integration_time != 12:
            res["Integration Time"] = integration_time
        res["Gain Setting"] = self._get_param("gainSetting")
        return res


class DsscConstants(enum.Enum):
    Offset = enum.auto()


class DsscCalcatFriend(BaseCalcatFriend):
    _constant_enum_class = DsscConstants

    def __init__(self, device, *args, **kwargs):
        super().__init__(device, *args, **kwargs)
        self._constants_need_conditions = {
            DsscConstants.Offset: self.dark_condition,
        }

    @staticmethod
    def add_schema(
        schema,
        managed_keys,
        param_prefix="constantParameters",
        status_prefix="foundConstants",
    ):
        super(DsscCalcatFriend, DsscCalcatFriend).add_schema(
            schema, managed_keys, "DSSC-Type", param_prefix, status_prefix
        )
        schema.setDefaultValue(f"{param_prefix}.memoryCells", 400)
        (
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.pulseIdChecksum")
            .assignmentOptional()
            .defaultValue(2.8866323107820637e-36)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.acquisitionRate")
            .assignmentOptional()
            .defaultValue(4.5)
            .commit(),
            DOUBLE_ELEMENT(schema)
            .key(f"{param_prefix}.encodedGain")
            .assignmentOptional()
            .defaultValue(67328)
            .commit(),
        )

        _add_status_schema_from_enum(schema, status_prefix, DsscConstants)

    def dark_condition(self):
        res = OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        # res["Pulse id checksum"] = self._get_param("pulseIdChecksum")
        # res["Acquisition rate"] = self._get_param("acquisitionRate")
        # res["Encoded gain"] = self._get_param("encodedGain")
        return res
